import logging
import os

import requests
from pennsieve2 import Pennsieve
from configparser import SectionProxy
from typing import List, Optional, Union
from ._default import ServiceBase


class PennsieveService(ServiceBase):
    """A wrapper for Pennsieve2 library

    Parameters:
    -----------
    config : dict
        A configuration with defined profile name (pennsieve_profile_name).
    connect : bool
        Determines if Sparc Client should initiate connection with Pennsieve Agent.

    Attributes:
    -----------
    default_headers : dict
        A dictionary with headers to make HTTP requests.
    host_api : str
        A default HTTP address of the Pennsieve.
    Pennsieve : object
        A class holding st


    Methods:
    --------
    connect()
        Establishes connection with Pennsieve Agent.
    info() -> str
        Returns the version of Pennsieve Agent.
    get_profile() -> str
        Returns the currently used profile.
    set_profile() -> str
        Changes the profile to the specified name.
    close() : None
        Closes Pennsieve Agent.
    list_datasets(...) : dict
        Returns a dictionary with datasets matching search criteria.
    list_files(...) : dict
        Returns a dictionary with datasets matching search criteria.
    list_filenames(...) : list
        Returns a dictionary with filenames stored at AWS matching search criteria.
    list_records(...) : dict
        Returns a dictionary with records matching search criteria.

    """

    default_headers = {
        "Content-Type": "application/json",
        "Accept": "application/json; charset=utf-8",
    }

    host_api = "https://api.pennsieve.io"
    Pennsieve: Pennsieve = None
    profile_name: str = None

    def __init__(
        self, config: Optional[Union[dict, SectionProxy]] = None, connect: bool = False
    ) -> None:
        logging.info("Initializing Pennsieve...")
        logging.debug(str(config))

        self.Pennsieve = Pennsieve(connect=False)
        if config is not None:
            self.profile_name = config.get("pennsieve_profile_name")
            logging.info("Profile: " + self.profile_name)
        else:
            logging.info("Profile: none")
        if connect:
            self.connect()  # profile_name=self.profile_name)

    def connect(self) -> Pennsieve:
        """Establishes connection with Pennsieve Agent."""
        logging.info("Connecting to Pennsieve...")

        if self.profile_name is not None:
            self.Pennsieve.connect(profile_name=self.profile_name)
        else:
            self.Pennsieve.connect()
        return self.Pennsieve

    def info(self) -> str:
        """Returns the version of Pennsieve Agent."""
        return self.Pennsieve.agent_version()

    def get_profile(self) -> str:
        """Returns currently used profile.

        Returns:
        --------
        A string with username.
        """
        return self.Pennsieve.get_user()

    def set_profile(self, profile_name: str) -> str:
        """Changes the profile to the specified name.

        Parameters:
        -----------
        profile_name : str
            The name of the profile to change into.

        Returns:
        --------
        A string with confirmation of profile switch.
        """
        return self.Pennsieve.switch(profile_name)

    def close(self) -> None:
        """Closes the Pennsieve Agent."""
        return self.Pennsieve.stop()

    def list_datasets(
        self,
        limit: int = 10,
        offset: int = 0,
        query: str = None,
        organization: str = None,
        organization_id: int = None,
        tags: List[str] = None,
        embargo: bool = None,
        order_by: str = None,
        order_direction: str = None,
    ) -> list:
        """Gets datasets matching specified criteria.

        Parameters:
        -----------
        limit : int
            max number of datasets returned
        offset : int
            offset used for pagination of results
        query : str
            fuzzy text search terms (refer to elasticsearch)
        organization : str
            publishing organization
        organization_id : int
            publishing organization id
        tags : list(str)
            match dataset tags
        embargo : bool
            include embargoed datasets
        order_by : str
            Field to order by:
                name - dataset name
                date - date published
                size - size of dataset
                relevance - order determined by elasticsearch
        order_direction : str
            Sort order:
                asc - Ascending, from A to Z
                desc - Descending, from Z to A

        Returns:
        --------
        A json with the results.

        """
        return self.Pennsieve.get(
            self.host_api + "/discover/search/datasets",
            headers=self.default_headers,
            params={
                "limit": limit,
                "offset": offset,
                "query": query,
                "organization": organization,
                "organizationId": organization_id,
                "tags": tags,
                "embargo": embargo,
                "orderBy": order_by,
                "orderDirection": order_direction,
            },
        )

    def list_files(
        self,
        limit: int = 10,
        offset: int = 0,
        file_type: str = None,
        query: str = None,
        organization: str = None,
        organization_id: int = None,
        dataset_id: int = None,
    ) -> list:
        """
        Parameters:
        -----------
        limit : int
            max number of datasets returned
        offset : int
            offset used for pagination of results
        file_type : str
            type of file
        query : str
            fuzzy text search terms (refer to elasticsearch)
        model : str
            only return records of this model
        organization : str
            publishing organization
        organization_id : int
            publishing organization id
        dataset_id : int
            files within this dataset

        Returns:
        --------
        List of files stored at AWS with their parameters.
        """

        return self.Pennsieve.get(
            self.host_api + "/discover/search/files",
            headers=self.default_headers,
            params={
                "limit": limit,
                "offset": offset,
                "fileType": file_type,
                "query": query,
                "organization": organization,
                "organizationId": organization_id,
                "datasetId": dataset_id,
            },
        )["files"]

    def list_filenames(
        self,
        limit: int = 10,
        offset: int = 0,
        file_type: str = None,
        query: str = None,
        organization: str = None,
        organization_id: int = None,
        dataset_id: int = None,
    ) -> list:
        """Calls list_files() and extracts the names of the files.
        See also
        --------
        list_files()
        """
        response = self.list_files(
            limit=limit,
            offset=offset,
            file_type=file_type,
            query=query,
            organization=organization,
            organization_id=organization_id,
            dataset_id=dataset_id,
        )

        return list(map(lambda x: "/".join(x["uri"].split("/")[5:]), response))

    def list_records(
        self,
        limit: int = 10,
        offset: int = 0,
        model: str = None,
        organization: str = None,
        dataset_id: int = None,
    ) -> list:
        """
        Parameters:
        -----------
        limit : int
            max number of datasets returned
        offset : int
            offset used for pagination of results
        model : str
            only return records of this model
        organization : str
            publishing organization
        dataset_id : int
            files within this dataset
        """

        return self.Pennsieve.get(
            self.host_api + "/discover/search/records",
            headers=self.default_headers,
            params={
                "limit": limit,
                "offset": offset,
                "model": model,
                "organization": organization,
                "datasetId": dataset_id,
            },
        )

    def download_file(self, file_list: dict, output_name: str = None):
        """Downloads files into a local storage.

        Parameters:
        -----------
        file_list : dict
            names of the file(s) to download with their parameters.
            The files need to come from a single database.
        output_name : str
            The name of the output file (used if the archive

        Returns:
        --------
        A response from the server.
        """

        # make sure we are passing a list
        file_list = [file_list] if type(file_list) is dict else file_list

        # create a tuple with datasetId and version of the dataset
        properties = set([(x["datasetId"], x["datasetVersion"]) for x in file_list])

        # extract all the files
        paths = [
            x if x.get("uri") is None else "/".join(x.get("uri").split("/")[5:]) for x in file_list
        ]
        assert (
            len(properties) == 1
        ), "Downloading files from multiple datasets or dataset versions is not supported."

        # initialize parameters for the request
        json = {
            "data": {
                "paths": paths,
                "datasetId": next(iter(properties))[0],
                "version": next(iter(properties))[1],
            }
        }

        # download the files with zipit service
        url = "https://api.pennsieve.io/zipit/discover"
        headers = {"content-type": "application/json"}
        response = requests.post(url, json=json, headers=headers)

        # replace extension of the file with '.gz' if downloading more than 1 file
        if output_name is None:
            output_name = (
                file_list[0]["name"] if len(paths) == 1 else os.path.splitext(file_list[0]) + ".gz"
            )
        with open(output_name, mode="wb+") as f:
            f.write(response.content)
        return response

    def get(self, url: str, **kwargs):
        """Invokes GET endpoint on a server. Passing server name in url is optional.

        Parameters:
        -----------
        url : str
            the address of the server endpoint to be called (e.g. api.pennsieve.io/datasets).
            The name of the server can be ommitted.
        kwargs : dict
            a dictionary storing additional information

        Returns:
        --------
        String in JSON format with response from the server.

        Example:
        --------
        p=Pennsieve()
        p.get('https://api.pennsieve.io/discover/datasets', params={'limit':20})

        """
        return self.Pennsieve.get(url, **kwargs)

    def post(self, url: str, json: dict, **kwargs):
        """Invokes POST endpoint on a server. Passing server name in url is optional.

        Parameters:
        -----------
        url : str
            the address of the server endpoint to be called (e.g. api.pennsieve.io/datasets).
            The name of the server can be omitted.
        json : dict
            a request payload with parameters defined by a given endpoint
        kwargs : dict
            additional information

        Returns:
        --------
        String in JSON format with response from the server.
        """
        return self.Pennsieve.post(url, json=json, **kwargs)

    def put(self, url: str, json: dict, **kwargs):
        """Invokes PUT endpoint on a server. Passing server name in url is optional.

        Parameters:
        -----------
        url : str
            the address of the server endpoint to be called (e.g. api.pennsieve.io/datasets).
            The name of the server can be omitted.
        json : dict
            a request payload with parameters defined by a given endpoint
        kwargs : dict
            additional information

        Returns:
        --------
        String in JSON format with response from the server.
        """
        return self.Pennsieve.put(url, json=json, **kwargs)

    def delete(self, url: str, **kwargs):
        """Invokes DELETE endpoint on a server. Passing server name in url is optional.

        Parameters:
        -----------
        url : str
            the address of the server endpoint to be called. The name of the server can be omitted.
        kwargs : dict
            additional information

        Returns:
        --------
        String in JSON format with response from the server.
        """
        return self.Pennsieve.delete(url, **kwargs)



import json
from pennsieve2 import Pennsieve

from sparc.client.services.pennsieve import PennsieveService


def test_connect_no_profile(mocker):
    expected = None
    mocker.patch("pennsieve2.Pennsieve.connect")
    p = PennsieveService(connect=False)
    pennsieve = p.connect()
    actual = pennsieve.get_user()
    assert actual == expected


def test_connect_false_with_profile(mocker, mock_user, mock_pennsieve):
    expected = "profile"
    mocker.patch("pennsieve2.Pennsieve.connect")
    mocker.patch("pennsieve2.Pennsieve.get_user", mock_user.get_user)
    mock_user.set_user("profile")
    p = PennsieveService(connect=False, config={"pennsieve_profile_name": "profile"})
    pennsieve = p.connect()
    actual = pennsieve.get_user()
    assert actual == expected


def test_connect_true_with_profile(mocker, mock_user, mock_pennsieve):
    expected = "test version"
    mocker.patch("pennsieve2.Pennsieve.connect", mock_pennsieve.connect)
    mocker.patch("pennsieve2.Pennsieve.agent_version", mock_pennsieve.agent_version)
    mocker.patch("pennsieve2.Pennsieve.get_user", mock_user.get_user)
    mock_user.set_user("profile")
    p = PennsieveService(connect=True, config={"pennsieve_profile_name": "profile"})
    pennsieve = p.connect()
    assert pennsieve is not None
    assert pennsieve.get_user() == "profile"
    assert pennsieve.agent_version() == expected


def test_get_profile(mocker, mock_pennsieve, mock_user):
    expected = "user"
    mocker.patch("pennsieve2.Pennsieve.get_user", mock_user.get_user)
    mock_user.set_user("user")

    p = PennsieveService(connect=False)
    actual = p.get_profile()
    assert actual == expected


def test_set_profile(mocker, mock_pennsieve):
    expected = "new user"
    mocker.patch("pennsieve2.Pennsieve.switch", mock_pennsieve.switch)

    class Manifest:
        manifest = "manifest"

        def __init__(self):
            pass

    p = PennsieveService(connect=False)
    p.manifest = Manifest()
    actual = p.set_profile("new user")
    assert expected == actual


def test_info(mocker, mock_pennsieve):
    expected = "test version"
    mocker.patch("pennsieve2.Pennsieve.agent_version", mock_pennsieve.agent_version)
    p = PennsieveService(connect=False)
    actual = p.info()
    assert expected == actual


def test_closed(mocker, mock_pennsieve):
    expected = "closed"
    mocker.patch("pennsieve2.Pennsieve.stop", mock_pennsieve.close)
    p = PennsieveService(connect=False)
    actual = p.close()
    assert expected == actual


def test_list_datasets(mocker, mock_pennsieve):
    expected = {
        "limit": 1,
        "offset": 0,
        "totalCount": 1,
        "datasets": [
            {
                "id": 2,
                "sourceDatasetId": 3,
                "name": "dataset",
                "description": "description",
                "ownerId": 4,
                "ownerFirstName": "John",
                "ownerLastName": "Smith",
                "ownerOrcid": "0000-0000-0000-0000",
                "organizationName": "organization",
                "organizationId": 1,
                "license": "license",
                "tags": ["tag1", "tag2"],
                "version": 5,
                "revision": None,
                "size": 6,
                "modelCount": [{"modelName": "model", "count": 0}],
                "fileCount": 1,
                "recordCount": 1,
                "uri": "s3://pennsieve/1/1/",
                "arn": "arn:aws:s3:::pennsieve/1/1/",
                "status": "PUBLISH_SUCCEEDED",
                "doi": "10.00000/abc123",
                "banner": "https://pennsieve/banner",
                "readme": "https://pennsieve/dataset-assets/1/1/readme.md",
                "contributors": [
                    {
                        "firstName": "John",
                        "middleInitial": None,
                        "lastName": "Smith",
                        "degree": None,
                        "orcid": "0000-0000-0000-0000",
                    }
                ],
                "collections": [{"id": 1, "name": "name"}],
                "externalPublications": [
                    {"doi": "10.0000/protocols.xx", "relationshipType": "IsSupplementedBy"}
                ],
                "sponsorship": {
                    "title": "title",
                    "imageUrl": "https://imageurl",
                    "markup": "markup",
                },
                "pennsieveSchemaVersion": "4.0",
                "createdAt": "2000-12-12",
                "updatedAt": "2000-12-12",
                "firstPublishedAt": "2000-12-12",
                "versionPublishedAt": "2000-12-12",
                "revisedAt": None,
                "embargo": False,
                "embargoReleaseDate": None,
                "embargoAccess": None,
            }
        ],
    }
    mocker.patch("pennsieve2.Pennsieve.get", mock_pennsieve.list_datasets)
    p = PennsieveService(connect=False)
    actual = p.list_datasets()
    assert expected == actual


def test_list_files(mocker, mock_pennsieve):
    expected = [
        {
            "name": "This is the filename.txt",
            "datasetId": 1,
            "datasetVersion": 1,
            "size": 1,
            "fileType": "TXT",
            "packageType": "package",
            "icon": "Icon",
            "uri": "s3://pennsieve/a/b/a.txt",
            "createdAt": None,
            "sourcePackageId": "N:package:aaaaaa",
        }
    ]

    mocker.patch("pennsieve2.Pennsieve.get", mock_pennsieve.list_files)
    p = PennsieveService(connect=False)
    actual = p.list_files()
    assert expected == actual


def test_list_filenames(mocker, mock_pennsieve):
    expected = ["a.txt"]
    mocker.patch("pennsieve2.Pennsieve.get", mock_pennsieve.list_files)
    p = PennsieveService(connect=False)
    actual = p.list_filenames({"uri": "http://pennsieve.io/dataset/file"})
    assert expected == actual


def test_list_records(mocker, mock_pennsieve):
    expected = {
        "limit": 10,
        "offset": 0,
        "totalCount": 10000,
        "records": [
            {
                "datasetId": 1,
                "version": 3,
                "model": "researcher",
                "properties": {
                    "hasORCIDId": "https://orcid.org/0000-0000-0000-0000",
                    "hasAffiliation": "UPenn",
                    "middleName": "A",
                    "hasRole": "",
                    "lastName": "Smith",
                    "firstName": "John",
                    "id": "aaaaaaa-aaaa-aaaa-aaaaaa",
                },
            }
        ],
    }

    mocker.patch("pennsieve2.Pennsieve.get", mock_pennsieve.list_records)
    p = PennsieveService(connect=False)
    actual = p.list_records()
    assert expected == actual


def test_download(mocker, mock_pennsieve):
    file_list = [
        {
            "name": "manifest.json",
            "datasetId": 1,
            "datasetVersion": 1,
            "size": 10000,
            "fileType": "Json",
            "packageType": "Unsupported",
            "icon": "JSON",
            "uri": "s3://pennsieve/1/2/manifest.json",
            "createdAt": None,
            "sourcePackageId": None,
        }
    ]

    class PennsieveResponse:
        status_code = 200
        content = b'{"content" : "content}'

    def response(url=None, json=None, headers=None):
        return PennsieveResponse()

    mocker.patch("requests.post", response)
    mocker.patch("os.open")
    mocker.patch("os.write")

    p = PennsieveService(connect=False)
    response = p.download_file(file_list=file_list)
    assert response.status_code == 200



import os.path

import pytest

from sparc.client import SparcClient


def test_class(config_file):
    c = SparcClient(connect=False, config_file=config_file)
    assert len(c.module_names) > 0


# Config file tests
def test_config_non_existing(config_file=None):
    with pytest.raises(RuntimeError):
        a = SparcClient(config_file, connect=False)


# Test config file with incorrect section pointer
def test_config_no_section(test_resources_dir):
    config_file = os.path.join(test_resources_dir, "dummy_config.ini")
    with pytest.raises(KeyError):
        a = SparcClient(config_file, connect=False)


def test_failed_add_module(config_file):
    client = SparcClient(connect=False, config_file=config_file)
    with pytest.raises(ModuleNotFoundError):
        client.add_module(paths="sparc.client.xyz", connect=False)


def test_add_module_connect(config_file):
    sc = SparcClient(config_file=config_file, connect=False)

    expected_module_config = {"module_param": "value"}
    sc.add_module("mock_service", config=expected_module_config, connect=True)

    assert "mock_service" in sc.module_names
    assert hasattr(sc, "mock_service")

    d = sc.mock_service
    from mock_service import MockService

    assert isinstance(d, MockService)
    assert d.init_connect_arg is True
    assert d.init_config_arg == expected_module_config
    assert d.connect_method_called is True


def test_add_pennsieve(config_file):
    sc = SparcClient(config_file=config_file, connect=False)
    assert "pennsieve" in sc.module_names
    assert hasattr(sc, "pennsieve")
    from sparc.client.services.pennsieve import PennsieveService

    assert isinstance(sc.pennsieve, PennsieveService)


def test_connect(config_file, monkeypatch):
    sc = SparcClient(config_file=config_file, connect=False)
    mock_connect_results = []

    def make_mock_connect(service_name):
        return lambda: mock_connect_results.append(service_name)

    for name in sc.module_names:
        service = getattr(sc, name)
        monkeypatch.setattr(service, "connect", make_mock_connect(name))

    sc.connect()
    assert mock_connect_results == sc.module_names



import logging
import os
from configparser import ConfigParser, SectionProxy
from importlib import import_module
from inspect import isabstract, isclass
from pathlib import Path
from pkgutil import iter_modules
from typing import Optional, Union

from .services import ServiceBase


class SparcClient(object):
    """
    The main class of the sparc.client library.

    This class is used to connect existing modules located in <projectbase>/services folder


    Parameters:
    -----------
    config_file : str
        The location of the file in INI format that is used to extract configuration variables.
        The config file needs to define a [global] section with the name of the default profile
        (in square brackets as well) which holds environmental variables used by the modules.
        Refer to configparser for further details.
    connect : bool (True)
        Calls connect() method of each of the modules.
        By default during initialization all modules are initialized and ready to be used,
        unless connect is set to False.


    Attributes:
    -----------
    module_names : list
        Stores the list of modules that are automatically loaded from the <projectbase>/services directory.

    Methods:
    --------
    add_module(path, config, connect):
        Adds and optionally connects to a module in a given path with configuration variables defined in config.
    connect()
        Connects all the modules by calling their connect() functions.
    """

    def __init__(self, config_file: str = "config/config.ini", connect: bool = True) -> None:
        # Read config file
        if not config_file:
            raise RuntimeError("Configuration file not given")

        config = ConfigParser()

        if os.path.isfile(config_file):
            config.read(config_file)
            logging.debug(str(config))
        current_config = config["global"]["default_profile"]

        logging.debug("Using the following config:")
        logging.debug(str(config[current_config]))
        self.module_names = []

        # iterate through the modules in the current package
        package_dir = os.path.join(Path(__file__).resolve().parent, "services")

        for _, module_name, _ in iter_modules([package_dir]):
            # import the module and iterate through its attributes
            self.add_module(
                f"{__package__}.services.{module_name}", config[current_config], connect
            )

    def add_module(
        self,
        paths: Union[str, list[str]],
        config: Optional[Union[dict, SectionProxy]] = None,
        connect: bool = True,
    ) -> None:
        """Adds and optionally connects to a module in a given path with configuration variables defined in config.

        Parameters:
        -----------
        paths : str or list[str]
            a path to the module
        config : dict or configparser.SectionProxy
            a dictionary (or Section of the config file parsed by ConfigParser) with the configuration variables
        connect : bool
            determines if the module should auto-connect
        """
        if not isinstance(paths, list):
            paths = [paths]

        for path in paths:
            module_name = path.split(".")[-1] if "." in path else path
            try:
                module = import_module(path)
                for attribute_name in dir(module):
                    attribute = getattr(module, attribute_name)
                    if (
                        isclass(attribute)
                        and issubclass(attribute, ServiceBase)
                        and not isabstract(attribute)
                    ):
                        # Add the class to this package's variables
                        self.module_names.append(module_name)
                        c = attribute(connect=connect, config=config)
                        setattr(self, module_name, c)
                        if connect:
                            c.connect()

            except ModuleNotFoundError:
                logging.debug(
                    "Skipping module. Failed to import from %s", f"{path=}", exc_info=True
                )
                raise

    def connect(self) -> bool:
        """Connects each of the modules loaded into self.module_names"""
        for module_name in self.module_names:
            module = getattr(self, module_name)
            if hasattr(module, "connect"):
                getattr(self, module_name).connect()
        return True



from .client import SparcClient

__all__: tuple[str, ...] = [
    "SparcClient",
    # "services.pennsieve.PennsieveService"
]


# import pkg_resources
# __version__ = pkg_resources.get_distribution("sparc.client").version



import json

import pytest
import requests
import responses
from requests.exceptions import HTTPError

from sparc.client.services import PennsieveService


@responses.activate
def test_register_responses():

    responses.add(
        method=responses.GET,
        url="http://api.pennsieve.io/datasets/",
        content_type="application/json",
        json={"key": "value"},
        status=200,
    )

    responses.add(
        method=responses.POST,
        url="http://api.pennsieve.io/datasets/1/versions/1/files/download-manifest",
        content_type="application/json",
        json={"key2": "value2"},
        status=200,
    )

    responses.add(
        method=responses.PUT,
        url="http://api.pennsieve.io/whatever",
        content_type="application/json",
        json={"key3": "value3"},
        status=200,
    )

    responses.add(
        method=responses.DELETE,
        url="http://api.pennsieve.io/whatever",
        content_type="application/json",
        json={"key4": "value4"},
        status=200,
    )

    p = PennsieveService(config={"pennsieve_profile_name": "test"}, connect=False)
    resp = p.get("http://api.pennsieve.io/datasets/")
    assert resp == {"key": "value"}

    resp = p.post(
        "http://api.pennsieve.io/datasets/1/versions/1/files/download-manifest",
        json=json.dumps({"datasetId": 1}),
    )
    assert resp == {"key2": "value2"}

    resp = p.put("http://api.pennsieve.io/whatever", json=json.dumps({"datasetId": 0}))
    assert resp == {"key3": "value3"}

    resp = p.delete("http://api.pennsieve.io/whatever", json=json.dumps({"datasetId": 0}))
    assert resp == {"key4": "value4"}
